
-- Shared parts for FNIP protocol drivers

    g_Common_Version = "1.2.10"
	
-- Used in: FNIP-12xPWM, FNIP-12xMZT, FNIP-12xIO, FNIP-RGB/010
-- changes 1.2.10
-- FNIP-12xIO Analog Value properties are updated when inputs close/open
-- FNIP-8x16Av4 support added
-- Variable "Online" added
-- changes in v1.2.9
-- ExecuteCommand moved from individual protocol drivers
-- Refresh status added to ExecuteCommand() and old bits have been deleted
-- changes in v1.2.8
-- hertbeat is sent even after g_Online has changed to false because heartbeat messages weren't replied multiple times
-- previously g_Online could get stuck in false state if the conncetion was temporarily down
-- for a shoter time than what C4 could detect so C4 would not attempt to reconnect so there was no chance
-- for g_Online to change to true so most functions remained blocked.
-- If the connection was down short enough, the heartbeats that weren't replied to
-- because they were sent after the connection got down but before the watchdog timer reached the value that set g_Online to false
-- were retransmitted by TCP/IP and could be replied to when the connection restored which reset the watchdog timer which
-- allowed g_Online to be set back to true.
-- with the heartbeat being sent regardless of the state of g_Online this happens even after TCP/IP retransmissions were given up on
-- the reason hertbeats are used is it takes C4 too long to detect connection loss  
-- CheckNetworkConnectionStatus() added and called in OnLateDriverInit and by a Lua command in the protocol drivers 
-- debug is set to Print by default so we can see what's happening right from the start 
-- v1.2.7
-- changes
-- sends connection status to all proxies when a GET_CONNECTED_STATE is received from proxy drivers
-- v1.2.6
-- changes
-- for TCP input status report both ch,status and status,channel orders are supported
-- driver and fw version checks removed
-- v1.2.5
-- changes
-- RGBBindings are now get updated about lost and restored connection with FN modules

g_debugprint=false
g_debuglog=false
g_DebugTimer=0


g_HeartBeatTimer=0
g_SendTimer = 0
g_ReadConfigTimer = 0
g_TickTimer = 0

g_SendBuffer=""
g_ReceiveBuffer=""

g_ServerBinding=6001
g_Port=7078
g_slow=false

g_Online=false


g_HoldTimers={}
g_DoubleClickTimers={}

g_NextDoubleClick={0,0,0,0,0,0,0,0,0,0,0,0}

g_TCPHeartBeatCountWOAns = 0

C4:AddVariable ("Online",0,"BOOL")
--C4:AddVariable ("Online","false","BOOL")
--C4:AddVariable ("Online","false","STRING")

function ReadConfig()
    local i = 0
    if(g_ReadConfigTimer~=0) then
	   g_ReadConfigTimer:Cancel()
    end

    if( ReadCustomConfigOnInit ~= nil ) then
	   ReadCustomConfigOnInit()
    end
    
    -- TODO: Move this part to the PWM's driver.lua file (Add to ReadCustomConfigOnInit function)
    if( g_Module_Type == "FNIP-12xPWM" ) then
	   if( C4:PersistGetValue("FNIP_LastRGBMode1") ~= nil) then
		  dbg("Reading 'LastRGBMode' from persistence.") 
		  for i=1,4 do
			 g_LastRGBMode[i] = C4:PersistGetValue("FNIP_LastRGBMode"..i)
		  end
	   else
		  print("No information stored about LastRGBMode" )
	   end
    end

    
    g_ReadConfigTimer = C4:SetTimer(500, function(timer)
	   i=i+1
	   dbg("[I] Read config["..i.."]: " .. g_ReadConfigCommands[i])
	   sendCommand(g_ReadConfigCommands[i])
	   if (i==#g_ReadConfigCommands) then
		  i=0
		  timer:Cancel()
	   end
    end,true)
end

function CheckNetworkConnectionStatus()
	id = C4:GetDeviceID()
	dbg ("my device id is " .. id)
	nwstat = C4:GetNetworkConnections()

	for k,v in pairs(nwstat) do
	 if(type(v)=="table") then
	   for t,y in pairs(v) do
		 if(type(y)=="table") then
			for p,q in pairs(y) do
			   if ((p == "deviceid") and (q == id)) then
				 dbg ("\nI am a " .. y.name .. "\nmy ip address is " .. y.address)
				 if (y.state == 1) then 
					dbg ("\nmy status is online")
					UpdateConnectionStatus(true)
				 elseif (y.state == 0) then
					dbg ("\nmy status is offline")
					UpdateConnectionStatus(false)
				 end--]]
			   end
			end
		  end
		end
	  end
	end
end

function OnDriverLateInit()	
    ReadConfig()
	CheckNetworkConnectionStatus()
	g_debugprint=true --so we can see what's happening right from the start
	C4:UpdateProperty("Debug Mode","Print")
	SetDebugTimer()
end

function OnSystemEvent(data)
    dbg("System Event occured - data: " .. data)
end

function OnConnectionStatusChanged(idBinding, nPort, strStatus)
    dbg("OnConnectionStatusChanged")
    if (idBinding == g_ServerBinding) then
	   if (strStatus == "ONLINE") then
		  UpdateConnectionStatus(true)
	   else
		  UpdateConnectionStatus(false)
	   end
    end
end

function ClearConnectionWatchdog()
    if (g_TCPHeartBeatCountWOAns>1 ) then
	   UpdateConnectionStatus(true)
    end
    g_TCPHeartBeatCountWOAns = 0
end

function UpdateConnectionStatus(online)
    SendToProxy(RGBBinding, "DYNAMIC_CAPABILITIES_CHANGED", { HAS_CONNECTION_STATUS = true })
    if (online) then
	   dbg("Online")
	   SetHeartBeatTimer()
	   SetTickTimer()
	   -- ClearDebugTimer()
	   ReadConfig()
	   --ClearConnectionWatchdog()
	   g_TCPHeartBeatCountWOAns = 0
	   C4:UpdateProperty("Module status","Online")
	   C4:SetVariable("Online",1)
	   SendToAllProxy("ONLINE_CHANGED", {STATE = "True"})
    else
	   --g_HeartBeatTimer:Cancel()
	   dbg("Offline")
	   C4:UpdateProperty("Module status","Offline")
	   C4:UpdateProperty("Module type","")
	   C4:UpdateProperty("MAC Address","")
	   C4:UpdateProperty("Firmware version","")
	   SendToAllProxy("ONLINE_CHANGED", {STATE = "False"})
	   C4:SetVariable("Online",0)
    end
    g_Online=online
end


-------------------------Debug and timers---------------------------------
function SetDebugTimer()	
    g_DebugTimer = C4:SetTimer(900000, function(timer) -- 15 minutes
	   dbg("Turning Debug Mode Off (timer expired)")
	   ClearDebugTimer()
    end,false)
end

function ClearDebugTimer()
    if (g_DebugTimer ~= 0) then g_DebugTimer:Cancel() end
    C4:UpdateProperty("Debug Mode", "Off")
    g_debugprint = false
    g_debuglog = false
end

function dbg(strDebugText)
    if (g_debugprint) then print(strDebugText) end
    if (g_debuglog) then C4:ErrorLog(strDebugText) end
end

function SetHeartBeatTimer()
    if( ( g_HeartBeatTimer == 0 ) or (g_HeartBeatTimer == nil)) then
	   g_HeartBeatTimer = C4:SetTimer(5000, function(timer)
		  if(g_Online) then
			 if ( g_TCPHeartBeatCountWOAns >= 3 ) then
				UpdateConnectionStatus(false)
			 end
			-- queueMessage("FN,TOG,1\r")
			-- queueMessage("HB\r")
			 g_TCPHeartBeatCountWOAns = g_TCPHeartBeatCountWOAns + 1
		 end
		queueMessage("HB\r")
		dbg("Online " .. tostring(g_Online))
		dbg("WatchdogCount " .. tostring (g_TCPHeartBeatCountWOAns))
		dbg("Queue " .. g_SendBuffer)
	   end,true)
    end
end

function SetTickTimer()
    g_TickTimer = C4:SetTimer(1000, function(timer)
	   if (g_Online) then
		  if( Tick ~= nil ) then
			 Tick()
		  end
	   else
		  timer:Cancel()
	   end
    end,true)
end



-- deprecated
function OnTimerExpired(idTimer)
    dbg("(W): Deprecated timer used")
    if(idTimer == g_SendTimer) then
	   g_SendTimer=C4:KillTimer(idTimer)
	   sendBufferedMessage()
    else
	   C4:KillTimer(idTimer)
    end
end


function SendToAllProxy(strCommand,tParams)
    for binding in ipairs(g_ChannelBindings) do
	   SendToProxy(g_ChannelBindings[binding], strCommand, tParams)
    end
	if (g_Module_Type == "FNIP-RGB/010") then
		SendToProxy(g_RGBBinding,strCommand,tParams)
	elseif ( g_Module_Type == "FNIP-12xPWM") then
		for binding in ipairs(g_RGBBindings) do
			SendToProxy(g_RGBBindings[binding], strCommand, tParams)
		end
	end
end


function SetSendTimer(duration)
    dbg("fn: SetSendTimer")
    g_SendTimer = C4:SetTimer(10, function(timer)
	   sendBufferedMessage()
	   --g_SendTimer=nil
    end,true)
end


function queueMessage(strMessage)
    dbg("Queueing Message: " .. strMessage)
    g_SendBuffer = g_SendBuffer..strMessage
    if( ( g_SendTimer == 0 ) or (g_SendTimer == nil)) then
	   SetSendTimer()
    end
end

function sendBufferedMessage()
    --if(g_Online) then
	   if (g_SendBuffer ~= "") then
		  if(string.len(g_SendBuffer)>15) then
			 tempBuffer=string.sub(g_SendBuffer,0,15)
			 g_SendBuffer=string.sub(g_SendBuffer,16,-1)
			 --[[if(g_SendBuffer~="") then
			 SetSendTimer(10)
			 --g_SendTimer=C4:AddTimer(10,"MILLISECONDS",true)
			 end]]--
		  else
			 tempBuffer=g_SendBuffer
			 g_SendBuffer=""
			 g_slow=false
		  end
		  dbg("Sending To Server: " .. tempBuffer)
		  C4:SendToNetwork(g_ServerBinding, g_Port, tempBuffer)
	   end
    --end
end

function sendCommand(strCommand)
    queueMessage("FN," .. strCommand .. "\r")
end

function ReceivedFromProxy(idBinding, strCommand, tParams)
    dbg("ReceivedFromProxy (" .. idBinding .. ", " .. strCommand .. ")")
	if(type(tParams)=="table") then
		for k,v in pairs(tParams) do
			dbg(k .. " : " .. tostring(v) .. "\n")
		end
	end
	if (strCommand == "GET_CONNECTED_STATE") then
		if (g_Online) then
			SendToAllProxy("ONLINE_CHANGED", {STATE = "True"})
		else
			SendToAllProxy("ONLINE_CHANGED", {STATE = "False"})
		end
	else
		CommandInterpreter(idBinding, strCommand, tParams)
	end
end

function SendToProxy(idBinding, strCommand, tParams, ...)
    if(idBinding==nil) then
	   return
    end
    dbg("SendToProxy (" .. idBinding .. ", " .. strCommand .. ")")
	if(type(tParams)=="table") then
		for k,v in pairs(tParams) do
			dbg(k .. " : " .. tostring(v) .. "\n")
		end
	end
    if ... then
	   local callType = ...
	   C4:SendToProxy(idBinding, strCommand, tParams, callType)
    else
	   C4:SendToProxy(idBinding, strCommand, tParams)
    end
end


function PropertyChangedCommon(strProperty)
    if (strProperty == "Debug Mode") then
	   if (Properties[strProperty] == "Off") then
		  ClearDebugTimer()
	   end
	   if (Properties[strProperty] == "Print") then
		  g_debugprint = true
		  g_debuglog = false
		  SetDebugTimer()
	   end
	   if (Properties[strProperty] == "Log") then
		  g_debugprint = false
		  g_debuglog = true
		  SetDebugTimer()
	   end
	   if (Properties[strProperty] == "Print and Log") then
		  g_debugprint = true
		  g_debuglog = true
		  SetDebugTimer()
	   end
    end
end


--- Command handling ---
function CommonCommandHandler(strDataTemp)
    if(strDataTemp~=nil) then
	   command=string.match(strDataTemp,"FN,(%a*)")
	   command=string.upper(command)
	   
	   
	   if(command=="IN") then
		  channel,param1=string.match(strDataTemp,"FN,IN,(%d+),(%a+)")
		  if( (param1 ~= "ON") and (param1 ~= "OFF") ) then
			 param1,channel=string.match(strDataTemp,"FN,IN,(%a+),(%d+)")
		  end

		  channel=tonumber(channel)
		  if(param1=="ON") then
			 if(channel~=nil) then
				if(g_input_id_shift~=nil) then
				    SendToProxy (g_input_id_shift+channel,"CLOSED",{},"NOTIFY")
				else
				    SendToProxy (20+channel,"CLOSED",{},"NOTIFY")
				end
				
				C4:FireEvent("Input".. channel .. " press")
				dbg("Input".. channel .. " press")

				g_HoldTimers[channel] = C4:SetTimer(500, function(timer)
				    C4:FireEvent("Input".. channel .. " hold")
				    dbg("Input".. channel .. " hold")
				    g_HoldTimers[channel] = nil
				    end)
				    
				if(g_DoubleClickTimers[channel]~=nil) then
				    g_DoubleClickTimers[channel]:Cancel()
				    g_DoubleClickTimers[channel]=nil
				    g_NextDoubleClick[channel]=1
				end
	--			if (g_Module_Type == "FNIP-12xIO") then 
	--			C4:UpdateProperty ("Analog Input ".. string.format("%02d", channel) .." Value" , "ON")
	--			end
			 end
		  elseif(param1=="OFF") then
			 if(channel~=nil) then
				if(g_input_id_shift~=nil) then
				    SendToProxy (g_input_id_shift+channel,"OPENED",{},"NOTIFY")
				else
				    SendToProxy (20+channel,"OPENED",{},"NOTIFY")
				end
				C4:FireEvent("Input".. channel .. " release")
				dbg("Input".. channel .. " release")

				if( g_UpDirections ~= nil ) then
				    if(g_switch_modes[channel]) then
					   --If input mode is toggle, switch the ramp direction
					   g_UpDirections[channel]=not g_UpDirections[channel]
				    end
				end
				
				-- if hold timer is running
				if(g_HoldTimers[channel]~=nil) then
				    g_HoldTimers[channel]:Cancel()
				    g_HoldTimers[channel]=nil
				    
				    if(g_NextDoubleClick[channel]==1) then
					   C4:FireEvent("Input".. channel .. " double tap")	
					   dbg("Input".. channel .. " double tap")
					   g_NextDoubleClick[channel]=0
				    else
					   g_DoubleClickTimers[channel]=C4:SetTimer(500, function(timer) -- TODO: Check MZT
						  C4:FireEvent("Input".. channel .. " single tap")
						  dbg("Input".. channel .. " single tap")
						  g_NextDoubleClick[channel]=0
						  g_DoubleClickTimers[channel] = nil
						  end)
				    end
				end
	--			if (g_Module_Type == "FNIP-12xIO") then 
	--				C4:UpdateProperty ("Analog Input ".. string.format("%02d", channel) .." Value" , "OFF")
	--			end
			 end
		  end
		  
		  
	   elseif(command=="VER") then
		  param1=string.match(strDataTemp,"FN,"..command..",(%w+%.%w+%.%w+)")
		  dbg("Firmware version: " .. param1)
		  C4:UpdateProperty("Firmware version",param1)
		  
	   elseif(command=="MAC") then
		  param1=string.match(strDataTemp,"FN,"..command..",(%w+:%w+:%w+:%w+:%w+:%w+)")
		  C4:UpdateProperty("MAC Address",param1)
		  
	   elseif(command=="MTYPE") then
		  param1=string.match(strDataTemp,"FN,"..command..",(%a*)")
		  
		  
		  if(param1 == "FNIP") then 
			 mtype=string.match(strDataTemp,"FN,"..command..",(%w+-%w+)")
			 if (mtype == "FNIP-RGB") then mtype = "FNIP-RGB/010" end
		  elseif(param1 == "SH") then 
			 mtype="FNIP-4xSH" 
		  elseif(param1 == "MZT") then 
			 mtype="FNIP-12xMZT" 
		  end
		  
		  
		  if (mtype == g_Module_Type)
		    or ( g_Module_Type == "FNIP-8x16A" and mtype == "FNIP-8x16Av4" )
		    or ( g_Module_Type == "FNIP-8x16A" and mtype == "FNIP-8x10A" )
			or ( g_Module_Type == "FNIP-8x10A" and mtype == "FNIP-8x16A" )then
			 C4:UpdateProperty("Module type",mtype)
		  else
			 C4:UpdateProperty("Module type","ERROR! Module type mismatch, please connect ".. g_Module_Type .." (" .. mtype .." has been recognized)")
		  end

	   else
		  dbg("(W): Unhandled command: " .. command)
	   end
    end
end


-- ExecuteCommand is used to handle commands from SendToDeviceAsync(), Actions, and Programming Commands
function ExecuteCommand(strCommand, tParams)
	dbg("ExecuteCommand called with : " .. strCommand)
	if (strCommand == "LUA_ACTION") then
		if tParams ~= nil then
			for cmd,cmdv in pairs(tParams) do 
				if cmd == "ACTION" then
				    if cmdv=="READ_SETTINGS" then
					   ReadConfig()
				    elseif cmdv =="REFRESH_STATUS" then
						CheckNetworkConnectionStatus()
					end
				end
			end
		end
	end
end

function tonumber_loc(str, base)
    if (type(str) == "number") then
        return str
    end
    local s = str:gsub(",", ".") -- Assume US Locale decimal separator
    local num = tonumber(s, base)
    if (num == nil) then
        s = str:gsub("%.", ",") -- Non-US Locale decimal separator
        num = tonumber(s, base)
    end
    return num
end
